@@ -4,12 +4,11 @@
 //! get a super-set of matches. Then, we we confirm each match using precise
 //! name resolution.
 
-use std::{convert::TryInto, mem};
+use std::{convert::TryInto, mem, sync::Arc};
 
 use base_db::{FileId, FileRange, SourceDatabase, SourceDatabaseExt};
 use hir::{
-    AsAssocItem, DefWithBody, HasAttrs, HasSource, InFile, ModuleDef, ModuleSource, Semantics,
-    Visibility,
+    AsAssocItem, DefWithBody, HasAttrs, HasSource, InFile, ModuleSource, Semantics, Visibility,
 };
 use once_cell::unsync::Lazy;
 use rustc_hash::FxHashMap;
@@ -58,13 +57,18 @@ impl IntoIterator for UsageSearchResult {
 pub struct FileReference {
     pub range: TextRange,
     pub name: ast::NameLike,
-    pub access: Option<ReferenceAccess>,
+    pub category: Option<ReferenceCategory>,
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub enum ReferenceAccess {
-    Read,
+pub enum ReferenceCategory {
+    // FIXME: Add this variant and delete the `retain_adt_literal_usages` function.
+    // Create
     Write,
+    Read,
+    // FIXME: Some day should be able to search in doc comments. Would probably
+    // need to switch from enum to bitflags then?
+    // DocComment
 }
 
 /// Generally, `search_scope` returns files that might contain references for the element.
@@ -212,13 +216,13 @@ impl Definition {
     fn search_scope(&self, db: &RootDatabase) -> SearchScope {
         let _p = profile::span("search_scope");
 
-        if let Definition::ModuleDef(hir::ModuleDef::BuiltinType(_)) = self {
+        if let Definition::BuiltinType(_) = self {
             return SearchScope::crate_graph(db);
         }
 
         // def is crate root
         // FIXME: We don't do searches for crates currently, as a crate does not actually have a single name
-        if let &Definition::ModuleDef(hir::ModuleDef::Module(module)) = self {
+        if let &Definition::Module(module) = self {
             if module.crate_root(db) == module {
                 return SearchScope::reverse_dependencies(db, module.krate());
             }
@@ -372,32 +376,63 @@ impl<'a> FindUsages<'a> {
             }
         };
 
-        let name = self.def.name(sema.db).or_else(|| {
-            self.include_self_kw_refs.as_ref().and_then(|ty| {
-                ty.as_adt()
-                    .map(|adt| adt.name(self.sema.db))
-                    .or_else(|| ty.as_builtin().map(|builtin| builtin.name()))
-            })
-        });
-        let name = match name {
-            Some(name) => name.to_string(),
+        let name = match self.def {
+            // special case crate modules as these do not have a proper name
+            Definition::Module(module) if module.crate_root(self.sema.db) == module => {
+                // FIXME: This assumes the crate name is always equal to its display name when it really isn't
+                module
+                    .krate()
+                    .display_name(self.sema.db)
+                    .map(|crate_name| crate_name.crate_name().as_smol_str().clone())
+            }
+            _ => {
+                let self_kw_refs = || {
+                    self.include_self_kw_refs.as_ref().and_then(|ty| {
+                        ty.as_adt()
+                            .map(|adt| adt.name(self.sema.db))
+                            .or_else(|| ty.as_builtin().map(|builtin| builtin.name()))
+                    })
+                };
+                self.def.name(sema.db).or_else(self_kw_refs).map(|it| it.to_smol_str())
+            }
+        };
+        let name = match &name {
+            Some(s) => s.as_str(),
             None => return,
         };
-        let name = name.as_str();
 
-        for (file_id, search_range) in search_scope {
-            let text = sema.db.file_text(file_id);
-            let search_range =
-                search_range.unwrap_or_else(|| TextRange::up_to(TextSize::of(text.as_str())));
-
-            let tree = Lazy::new(|| sema.parse(file_id).syntax().clone());
-
-            for (idx, _) in text.match_indices(name) {
+        // these can't be closures because rust infers the lifetimes wrong ...
+        fn match_indices<'a>(
+            text: &'a str,
+            name: &'a str,
+            search_range: TextRange,
+        ) -> impl Iterator<Item = TextSize> + 'a {
+            text.match_indices(name).filter_map(move |(idx, _)| {
                 let offset: TextSize = idx.try_into().unwrap();
                 if !search_range.contains_inclusive(offset) {
-                    continue;
+                    return None;
                 }
+                Some(offset)
+            })
+        }
+        fn scope_files<'a>(
+            sema: &'a Semantics<RootDatabase>,
+            scope: &'a SearchScope,
+        ) -> impl Iterator<Item = (Arc<String>, FileId, TextRange)> + 'a {
+            scope.entries.iter().map(|(&file_id, &search_range)| {
+                let text = sema.db.file_text(file_id);
+                let search_range =
+                    search_range.unwrap_or_else(|| TextRange::up_to(TextSize::of(text.as_str())));
 
+                (text, file_id, search_range)
+            })
+        }
+
+        for (text, file_id, search_range) in scope_files(sema, &search_scope) {
+            let tree = Lazy::new(move || sema.parse(file_id).syntax().clone());
+
+            // Search for occurrences of the items name
+            for offset in match_indices(&text, name, search_range) {
                 for name in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                     if match name {
                         ast::NameLike::NameRef(name_ref) => self.found_name_ref(&name_ref, sink),
@@ -408,13 +443,9 @@ impl<'a> FindUsages<'a> {
                     }
                 }
             }
+            // Search for occurrences of the `Self` referring to our type
             if let Some(self_ty) = &self.include_self_kw_refs {
-                for (idx, _) in text.match_indices("Self") {
-                    let offset: TextSize = idx.try_into().unwrap();
-                    if !search_range.contains_inclusive(offset) {
-                        continue;
-                    }
-
+                for offset in match_indices(&text, "Self", search_range) {
                     for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                         if self.found_self_ty_name_ref(self_ty, &name_ref, sink) {
                             return;
@@ -424,9 +455,40 @@ impl<'a> FindUsages<'a> {
             }
         }
 
+        // Search for `super` and `crate` resolving to our module
+        match self.def {
+            Definition::Module(module) => {
+                let scope = search_scope.intersection(&SearchScope::module(self.sema.db, module));
+
+                let is_crate_root = module.crate_root(self.sema.db) == module;
+
+                for (text, file_id, search_range) in scope_files(sema, &scope) {
+                    let tree = Lazy::new(move || sema.parse(file_id).syntax().clone());
+
+                    for offset in match_indices(&text, "super", search_range) {
+                        for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
+                            if self.found_name_ref(&name_ref, sink) {
+                                return;
+                            }
+                        }
+                    }
+                    if is_crate_root {
+                        for offset in match_indices(&text, "crate", search_range) {
+                            for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
+                                if self.found_name_ref(&name_ref, sink) {
+                                    return;
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+            _ => (),
+        }
+
         // search for module `self` references in our module's definition source
         match self.def {
-            Definition::ModuleDef(hir::ModuleDef::Module(module)) if self.search_self_mod => {
+            Definition::Module(module) if self.search_self_mod => {
                 let src = module.definition_source(sema.db);
                 let file_id = src.file_id.original_file(sema.db);
                 let (file_id, search_range) = match src.value {
@@ -435,18 +497,25 @@ impl<'a> FindUsages<'a> {
                     ModuleSource::SourceFile(_) => (file_id, None),
                 };
 
+                let search_range = if let Some(&range) = search_scope.entries.get(&file_id) {
+                    match (range, search_range) {
+                        (None, range) | (range, None) => range,
+                        (Some(range), Some(search_range)) => match range.intersect(search_range) {
+                            Some(range) => Some(range),
+                            None => return,
+                        },
+                    }
+                } else {
+                    return;
+                };
+
                 let text = sema.db.file_text(file_id);
                 let search_range =
                     search_range.unwrap_or_else(|| TextRange::up_to(TextSize::of(text.as_str())));
 
                 let tree = Lazy::new(|| sema.parse(file_id).syntax().clone());
 
-                for (idx, _) in text.match_indices("self") {
-                    let offset: TextSize = idx.try_into().unwrap();
-                    if !search_range.contains_inclusive(offset) {
-                        continue;
-                    }
-
+                for offset in match_indices(&text, "self", search_range) {
                     for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                         if self.found_self_module_name_ref(&name_ref, sink) {
                             return;
@@ -472,7 +541,7 @@ impl<'a> FindUsages<'a> {
                 let reference = FileReference {
                     range,
                     name: ast::NameLike::NameRef(name_ref.clone()),
-                    access: None,
+                    category: None,
                 };
                 sink(file_id, reference)
             }
@@ -486,12 +555,12 @@ impl<'a> FindUsages<'a> {
         sink: &mut dyn FnMut(FileId, FileReference) -> bool,
     ) -> bool {
         match NameRefClass::classify(self.sema, name_ref) {
-            Some(NameRefClass::Definition(def @ Definition::ModuleDef(_))) if def == self.def => {
+            Some(NameRefClass::Definition(def @ Definition::Module(_))) if def == self.def => {
                 let FileRange { file_id, range } = self.sema.original_range(name_ref.syntax());
                 let reference = FileReference {
                     range,
                     name: ast::NameLike::NameRef(name_ref.clone()),
-                    access: None,
+                    category: None,
                 };
                 sink(file_id, reference)
             }
@@ -510,7 +579,7 @@ impl<'a> FindUsages<'a> {
                 let reference = FileReference {
                     range,
                     name: ast::NameLike::Lifetime(lifetime.clone()),
-                    access: None,
+                    category: None,
                 };
                 sink(file_id, reference)
             }
@@ -529,7 +598,7 @@ impl<'a> FindUsages<'a> {
                 let reference = FileReference {
                     range,
                     name: ast::NameLike::NameRef(name_ref.clone()),
-                    access: reference_access(&def, name_ref),
+                    category: ReferenceCategory::new(&def, name_ref),
                 };
                 sink(file_id, reference)
             }
@@ -539,7 +608,7 @@ impl<'a> FindUsages<'a> {
                     let reference = FileReference {
                         range,
                         name: ast::NameLike::NameRef(name_ref.clone()),
-                        access: reference_access(&def, name_ref),
+                        category: ReferenceCategory::new(&def, name_ref),
                     };
                     sink(file_id, reference)
                 } else {
@@ -550,14 +619,19 @@ impl<'a> FindUsages<'a> {
                 let field = Definition::Field(field);
                 let FileRange { file_id, range } = self.sema.original_range(name_ref.syntax());
                 let access = match self.def {
-                    Definition::Field(_) if field == self.def => reference_access(&field, name_ref),
+                    Definition::Field(_) if field == self.def => {
+                        ReferenceCategory::new(&field, name_ref)
+                    }
                     Definition::Local(l) if local == l => {
-                        reference_access(&Definition::Local(local), name_ref)
+                        ReferenceCategory::new(&Definition::Local(local), name_ref)
                     }
                     _ => return false,
                 };
-                let reference =
-                    FileReference { range, name: ast::NameLike::NameRef(name_ref.clone()), access };
+                let reference = FileReference {
+                    range,
+                    name: ast::NameLike::NameRef(name_ref.clone()),
+                    category: access,
+                };
                 sink(file_id, reference)
             }
             _ => false,
@@ -580,41 +654,41 @@ impl<'a> FindUsages<'a> {
                     range,
                     name: ast::NameLike::Name(name.clone()),
                     // FIXME: mutable patterns should have `Write` access
-                    access: Some(ReferenceAccess::Read),
+                    category: Some(ReferenceCategory::Read),
                 };
                 sink(file_id, reference)
             }
             Some(NameClass::ConstReference(def)) if self.def == def => {
                 let FileRange { file_id, range } = self.sema.original_range(name.syntax());
-                let reference =
-                    FileReference { range, name: ast::NameLike::Name(name.clone()), access: None };
+                let reference = FileReference {
+                    range,
+                    name: ast::NameLike::Name(name.clone()),
+                    category: None,
+                };
                 sink(file_id, reference)
             }
             // Resolve trait impl function definitions to the trait definition's version if self.def is the trait definition's
-            Some(NameClass::Definition(Definition::ModuleDef(mod_def))) => {
+            Some(NameClass::Definition(def)) if def != self.def => {
                 /* poor man's try block */
                 (|| {
-                    let this = match self.def {
-                        Definition::ModuleDef(this) if this != mod_def => this,
-                        _ => return None,
-                    };
-                    let this_trait = this
+                    let this_trait = self
+                        .def
                         .as_assoc_item(self.sema.db)?
                         .containing_trait_or_trait_impl(self.sema.db)?;
-                    let trait_ = mod_def
+                    let trait_ = def
                         .as_assoc_item(self.sema.db)?
                         .containing_trait_or_trait_impl(self.sema.db)?;
-                    (trait_ == this_trait
-                        && self.def.name(self.sema.db) == mod_def.name(self.sema.db))
-                    .then(|| {
-                        let FileRange { file_id, range } = self.sema.original_range(name.syntax());
-                        let reference = FileReference {
-                            range,
-                            name: ast::NameLike::Name(name.clone()),
-                            access: None,
-                        };
-                        sink(file_id, reference)
-                    })
+                    (trait_ == this_trait && self.def.name(self.sema.db) == def.name(self.sema.db))
+                        .then(|| {
+                            let FileRange { file_id, range } =
+                                self.sema.original_range(name.syntax());
+                            let reference = FileReference {
+                                range,
+                                name: ast::NameLike::Name(name.clone()),
+                                category: None,
+                            };
+                            sink(file_id, reference)
+                        })
                 })()
                 .unwrap_or(false)
             }
@@ -625,49 +699,48 @@ impl<'a> FindUsages<'a> {
 
 fn def_to_ty(sema: &Semantics<RootDatabase>, def: &Definition) -> Option<hir::Type> {
     match def {
-        Definition::ModuleDef(def) => match def {
-            ModuleDef::Adt(adt) => Some(adt.ty(sema.db)),
-            ModuleDef::TypeAlias(it) => Some(it.ty(sema.db)),
-            ModuleDef::BuiltinType(it) => {
-                let graph = sema.db.crate_graph();
-                let krate = graph.iter().next()?;
-                let root_file = graph[krate].root_file_id;
-                let module = sema.to_module_def(root_file)?;
-                Some(it.ty(sema.db, module))
-            }
-            _ => None,
-        },
+        Definition::Adt(adt) => Some(adt.ty(sema.db)),
+        Definition::TypeAlias(it) => Some(it.ty(sema.db)),
+        Definition::BuiltinType(it) => {
+            let graph = sema.db.crate_graph();
+            let krate = graph.iter().next()?;
+            let root_file = graph[krate].root_file_id;
+            let module = sema.to_module_def(root_file)?;
+            Some(it.ty(sema.db, module))
+        }
         Definition::SelfType(it) => Some(it.self_ty(sema.db)),
         _ => None,
     }
 }
 
-fn reference_access(def: &Definition, name_ref: &ast::NameRef) -> Option<ReferenceAccess> {
-    // Only Locals and Fields have accesses for now.
-    if !matches!(def, Definition::Local(_) | Definition::Field(_)) {
-        return None;
-    }
+impl ReferenceCategory {
+    fn new(def: &Definition, r: &ast::NameRef) -> Option<ReferenceCategory> {
+        // Only Locals and Fields have accesses for now.
+        if !matches!(def, Definition::Local(_) | Definition::Field(_)) {
+            return None;
+        }
 
-    let mode = name_ref.syntax().ancestors().find_map(|node| {
+        let mode = r.syntax().ancestors().find_map(|node| {
         match_ast! {
-            match (node) {
+            match node {
                 ast::BinExpr(expr) => {
                     if matches!(expr.op_kind()?, ast::BinaryOp::Assignment { .. }) {
                         // If the variable or field ends on the LHS's end then it's a Write (covers fields and locals).
                         // FIXME: This is not terribly accurate.
                         if let Some(lhs) = expr.lhs() {
-                            if lhs.syntax().text_range().end() == name_ref.syntax().text_range().end() {
-                                return Some(ReferenceAccess::Write);
+                            if lhs.syntax().text_range().end() == r.syntax().text_range().end() {
+                                return Some(ReferenceCategory::Write);
                             }
                         }
                     }
-                    Some(ReferenceAccess::Read)
+                    Some(ReferenceCategory::Read)
                 },
                 _ => None
             }
         }
     });
 
-    // Default Locals and Fields to read
-    mode.or(Some(ReferenceAccess::Read))
+        // Default Locals and Fields to read
+        mode.or(Some(ReferenceCategory::Read))
+    }
 }
